import semver
import yaml

from . import __version__, env


def get_schema_definition() -> dict:
    """
    Look up and read the current schema definition
    :return The schema definition
    :rtype dict
    """

    path = env.SCHEMA_DEFINITION_FILE
    with open(path) as fp:
        return yaml.load(fp, Loader=yaml.FullLoader)


def get_current_schema_version() -> str:
    current_version: semver.Version = semver.Version.parse(__version__)
    return f"{str(current_version.major)}.{str(current_version.minor)}.0"



import logging
import traceback
from typing import Dict, List, Optional

import pandas as pd
from cellxgene_schema import gencode
from cellxgene_schema.env import SCHEMA_REFERENCE_BASE_URL, SCHEMA_REFERENCE_FILE_NAME
from cellxgene_schema.validate import ONTOLOGY_PARSER, Validator

from .utils import get_hash_digest_column, getattr_anndata

logger = logging.getLogger(__name__)


class AnnDataLabelAppender:
    """
    From valid h5ad, handles writing a new h5ad file with ontology/gene labels added
    to adata.obs and adata.var respectively as indicated in the schema definition
    """

    def __init__(self, validator: Validator):
        """
        From a list of ids and defined constraints, creates a mapping dictionary {id: label, ...}

        :param Validator validator: a Validator object, it's used to get adata and schema defintion for its validation,
        it's also used to make sure the validation on this adata was successful.
        """

        if not validator.is_valid:
            raise ValueError(
                "AnnData object is not valid or hasn't been run through validation. "
                "Validate AnnData first before attempting to write labels"
            )
        self.adata = validator.adata
        self.validator = validator
        self.schema_def = validator.schema_def
        self.errors = []
        self.was_writing_successful = False

    def _merge_dicts(self, dict1: dict, dict2: dict) -> dict:
        """
        Recursively merges two dicts, designed to be used to flatten a column definition.

        :params dict dict1: first dict
        :params dict dict2: second dict

        :rtype dict
        :return the merged dict
        """

        merged_dict = dict1.copy()

        for key, value_2 in dict2.items():
            if key == "rule":
                continue

            if key not in dict1:
                merged_dict[key] = value_2
            else:
                value_1 = dict1[key]

                if not type(value_1) == type(value_2):
                    raise ValueError("Inconsistent types, impossible to merge")

                if isinstance(value_2, str):
                    if key == "error_message_suffix":
                        merged_dict[key] = value_1 + " " + value_2
                        continue
                    if not value_2 == value_1:
                        raise ValueError(f"Strings types in dependencies cannot be different, {value_1} and {value_2}")

                elif isinstance(value_2, list):
                    merged_dict[key] = list(set(value_1 + value_2))
                elif isinstance(value_2, dict):
                    merged_dict[key] = self._merge_dicts(value_1, value_2)
                else:
                    raise ValueError(f"merging {type(value_2)} is not implemented")

        return merged_dict

    def _flatten_column_def_with_dependencies(self, column_def: dict) -> dict:
        """
        Flattens a column definition  that has dependencies, it essentially concatenates all the definitions of the
        dependencies into the on definition

        :params dict column_def: the column definition that has dependencies in it

        :rtype dict
        :return the flatten column definition
        """

        # Do nothing if ther are no dependencies
        if "dependencies" not in column_def:
            return column_def

        flatten = column_def.copy()
        del flatten["dependencies"]

        for dep in column_def["dependencies"]:
            flatten = self._merge_dicts(flatten, dep)

        return flatten

    def _get_ontology_term_label(self, term_id: str, allowed_ontologies: List[str]) -> str:
        """
        Fetches human-readable label corresponding to a single ontology term_id, if it is a term_id defined in one of the
         allowed_ontologies. Raises ValueError if no ontology term label is found (should've triggered validation
         error if term_id is not valid).

        :param term_id: str single ontology term ID
        :param allowed_ontologies: List[str] list of ontologies to check for term_id label in
        :return: str term label
        """
        for ontology_name in allowed_ontologies:
            if ontology_name == "NA":
                continue
            elif ONTOLOGY_PARSER.is_valid_term_id(term_id, ontology_name):
                return ONTOLOGY_PARSER.get_term_label(term_id)
        raise ValueError(f"Add labels error: Unable to get label for '{term_id}'")

    def _get_mapping_dict_curie(self, ids: List[str], curie_constraints: dict) -> Dict[str, str]:
        """
        From defined constraints it creates a mapping dictionary of ontology IDs and labels.

        :param list[str] ids: Ontology IDs use for mapping
        :param list[str] curie_constraints: curie constraints e.g.
        schema_def["components"]["obs"]["columns"]["cell_type_ontology_term_id"]["curie_"]

        :return a mapping dictionary: {id: label, ...}
        :rtype dict
        """

        mapping_dict = {}
        allowed_ontologies = curie_constraints["ontologies"]
        multi_term_def = curie_constraints.get("multi_term")
        delimiter = None if multi_term_def is None else multi_term_def["delimiter"]

        # Map term_ids to their human-readable ontology labels
        for term_id in ids:
            # If there are exceptions the label should be the same as the id
            if "exceptions" in curie_constraints and term_id in curie_constraints["exceptions"]:
                mapping_dict[term_id] = term_id
                continue

            if delimiter is not None:
                labels = [self._get_ontology_term_label(term, allowed_ontologies) for term in term_id.split(delimiter)]
                mapping_dict[term_id] = delimiter.join(labels)
            else:
                mapping_dict[term_id] = self._get_ontology_term_label(term_id, allowed_ontologies)

        return mapping_dict

    def _get_mapping_dict_feature_id(self, ids: List[str]) -> Dict[str, str]:
        """
        Creates a mapping dictionary of gene/feature IDs and labels.

        :param list[str] ids: Gene/feature IDs use for mapping

        :return a mapping dictionary: {id: label, ...}
        :rtype dict
        """

        mapping_dict = {}

        for i in ids:
            organism = gencode.get_organism_from_feature_id(i)
            mapping_dict[i] = self.validator.gene_checkers[organism].get_symbol(i)

        return mapping_dict

    def _get_mapping_dict_feature_reference(self, ids: List[str]) -> Dict[str, Optional[gencode.SupportedOrganisms]]:
        """
        Creates a mapping dictionary of gene/feature IDs and NCBITaxon curies

        :param list[str] ids: Gene/feature IDs use for mapping

        :return a mapping dictionary: {id: label, ...}
        :rtype dict
        """

        mapping_dict = {}

        for i in ids:
            organism = gencode.get_organism_from_feature_id(i)
            mapping_dict[i] = organism.value

        return mapping_dict

    def _get_mapping_dict_feature_type(self, ids: List[str]) -> Dict[str, str]:
        """
        Creates a mapping dictionary of gene/feature IDs and its feature type.

        :param list[str] ids: Gene/feature IDs use for mapping

        :return a mapping dictionary: {id: feature_type, ...}
        :rtype dict
        """

        mapping_dict = {}

        for i in ids:
            organism = gencode.get_organism_from_feature_id(i)
            mapping_dict[i] = self.validator.gene_checkers[organism].get_type(i)

        return mapping_dict

    def _get_mapping_dict_feature_biotype(self, ids: List[str]) -> Dict[str, str]:
        """
        Creates a mapping dictionary of feature IDs and biotype ("gene" or "spike-in")

        :param list[str] ids: feature IDs use for mapping

        :return a mapping dictionary: {id: "gene", id: "spike-in", ...}
        :rtype dict
        """
        mapping_dict = {}

        for i in ids:
            if i.startswith("ERCC"):
                mapping_dict[i] = "spike-in"
            elif i.startswith("ENS"):
                mapping_dict[i] = "gene"
            else:
                raise ValueError(f"{i} is not a recognized `feature_name` and cannot be assigned a `feature_type`")

        return mapping_dict

    def _get_mapping_dict_feature_length(self, ids: List[str]) -> Dict[str, int]:
        """
        Creates a mapping dictionary of feature IDs and feature length, fetching from pre-calculated gene info CSVs
        derived from GENCODE mappings for supported organisms. Set to 0 for non-gene features.

        :param list[str] ids: feature IDs use for mapping

        :return a mapping dictionary: {id: <int>, id: 0, ...}
        :rtype dict
        """
        mapping_dict = {}

        for i in ids:
            organism = gencode.get_organism_from_feature_id(i)
            mapping_dict[i] = self.validator.gene_checkers[organism].get_length(i)

        return mapping_dict

    def _get_labels(
        self,
        component: str,
        column: str,
        column_definition: dict,
        label_type: dict,
    ) -> pd.Series:
        """
        Retrieves a new column (pandas categorical) with labels based on the IDs in 'column' and the logic in the
        'column_definition'

        :param str component: what dataframe in self.adata to work with
        :param str column: Column in self.adata with IDs that will be used to retrieve values
        :param dict column_definition: schema definition of the column
        e.g. schema_def["obs"]["columns"]["cell_type_ontology_term_id"]
        :param dict label_type: the type of label

        :rtype pandas.Categorical
        :return new pandas column with labels corresponding to input column
        """

        # Set variables for readability
        current_df = getattr_anndata(self.adata, component)

        if column == "index":
            original_column = pd.Series(current_df.index)
            original_column.index = current_df.index
        else:
            original_column = getattr(current_df, column)

        ids = getattr(current_df, column).drop_duplicates().tolist()

        # Flatten column definition (will do so if there are dependencies in the definition
        column_definition = self._flatten_column_def_with_dependencies(column_definition)

        if label_type == "curie":
            if "curie_constraints" not in column_definition:
                raise ValueError(
                    f"Schema definition error: 'add_labels' with type 'curie' was found for '{column}' "
                    "but no curie constraints were found for the labels"
                )

            mapping_dict = self._get_mapping_dict_curie(ids, column_definition["curie_constraints"])

        elif label_type == "feature_id":
            mapping_dict = self._get_mapping_dict_feature_id(ids=ids)

        elif label_type == "feature_reference":
            mapping_dict = self._get_mapping_dict_feature_reference(ids=ids)

        elif label_type == "feature_biotype":
            mapping_dict = self._get_mapping_dict_feature_biotype(ids=ids)

        elif label_type == "feature_length":
            mapping_dict = self._get_mapping_dict_feature_length(ids=ids)

        elif label_type == "feature_type":
            mapping_dict = self._get_mapping_dict_feature_type(ids=ids)

        else:
            raise TypeError(f"'{label_type}' is not supported in 'add-labels' functionality")

        new_column = original_column.copy().map(mapping_dict).astype("category")

        return new_column

    def _add_column(self, component: str, column: str, column_definition: dict):
        """
        Adds a new column (pandas categorical) to a component of adata with labels based on the IDs
        in 'column' and the logic in the 'column_def'

        :param str component: what dataframe in self.adata to work with
        :param str column: Column in self.adata with IDs that will be used to retrieve values
        :param dict column_definition: schema definition of the column
        e.g. schema_def["obs"]["columns"]["cell_type_ontology_term_id"]

        :rtype None
        """

        for label_def in column_definition["add_labels"]:
            new_column = self._get_labels(component, column, column_definition, label_def["type"])
            new_column_name = label_def["to_column"]
            getattr_anndata(self.adata, component)[new_column_name] = new_column

    def _add_labels(self):
        """
        Add columns to dataset dataframes based on values in other columns, as defined in schema definition yaml.
        """
        for component in ["obs", "var", "raw.var"]:
            # If the component does not exist, skip (this is for raw.var)
            if getattr_anndata(self.adata, component) is None:
                continue

            # Doing it for columns
            if "columns" in self.schema_def["components"][component]:
                for column, column_def in self.schema_def["components"][component]["columns"].items():
                    if "add_labels" in column_def:
                        self._add_column(component, column, column_def)

            # Doing it for index
            index_def = self.schema_def["components"][component]["index"]
            if "add_labels" in index_def:
                self._add_column(component, "index", index_def)

    def _remove_categories_with_zero_values(self):
        df = self.adata.obs
        for column in df.columns:
            col = df[column]
            if col.dtype == "category":
                df[column] = col.cat.remove_unused_categories()

    def _build_schema_reference_url(self, schema_version: str):
        return f"{SCHEMA_REFERENCE_BASE_URL}/{schema_version}/{SCHEMA_REFERENCE_FILE_NAME}"

    def write_labels(self, add_labels_file: str):
        """
        From a valid (per cellxgene's schema) h5ad, this function writes a new h5ad file with ontology/gene labels added
        to adata.obs  and adata.var respectively

        :param str add_labels_file: Path to new h5ad file with ontology/gene labels added

        :rtype None
        """
        logger.info("Writing labels")
        # Add columns to dataset dataframes based on values in other columns, as defined in schema definition yaml
        self._add_labels()

        # Remove unused categories
        self._remove_categories_with_zero_values()

        # Annotate Reserved Columns

        self.adata.uns["schema_version"] = self.validator.schema_version
        self.adata.uns["schema_reference"] = self._build_schema_reference_url(self.validator.schema_version)
        self.adata.obs["observation_joinid"] = get_hash_digest_column(self.adata.obs)

        # Write file
        try:
            self.adata.write_h5ad(add_labels_file, compression="gzip")
        except Exception as e:
            tb = traceback.format_exc()
            self.errors.append((f"Writing h5ad was unsuccessful, got exception '{e}'.", tb))

        # Print errors if any
        if self.errors:
            for e, tb in self.errors:
                logger.error(e, extra={"exec_info": tb})
            self.was_writing_successful = False
        else:
            self.was_writing_successful = True



import logging
import os
import sys
from base64 import b85encode
from typing import Dict, List, Union

import anndata as ad
import numpy as np
from scipy import sparse
from xxhash import xxh3_64_intdigest

logger = logging.getLogger(__name__)

SPARSE_MATRIX_TYPES = {"csc", "csr", "coo"}


def replace_ontology_term(dataframe, ontology_name, update_map):
    column_name = f"{ontology_name}_ontology_term_id"
    if dataframe[column_name].dtype != "category":
        dataframe[column_name] = dataframe[column_name].astype("category")
    for old_term, new_term in update_map.items():
        if old_term in dataframe[column_name].cat.categories:
            # add new one if not already in category, else continue
            if new_term not in dataframe[column_name].cat.categories:
                dataframe[column_name] = dataframe[column_name].cat.add_categories(new_term)
            # replace in dataset
            dataframe.loc[dataframe[column_name] == old_term, column_name] = new_term
            # remove deprecated_term from category
            dataframe[column_name] = dataframe[column_name].cat.remove_categories(old_term)


def map_ontology_term(dataframe, ontology_name, map_from_column, update_map):
    column_name = f"{ontology_name}_ontology_term_id"
    if dataframe[column_name].dtype != "category":
        dataframe[column_name] = dataframe[column_name].astype("category")
    for map_value, new_term in update_map.items():
        if new_term not in dataframe[column_name].cat.categories:
            dataframe[column_name] = dataframe[column_name].cat.add_categories(new_term)
        dataframe.loc[dataframe[map_from_column] == map_value, column_name] = new_term
    dataframe[column_name] = dataframe[column_name].cat.remove_unused_categories()


def remove_deprecated_features(*, adata: ad.AnnData, deprecated: List[str]) -> ad.AnnData:
    # Filter out genes that don't appear in the approved annotation
    var_to_keep = adata.var.index[~adata.var.index.isin(deprecated)].tolist()
    adata = adata[:, var_to_keep]

    # Repeat much of the same steps for the raw.var, if it exists
    if adata.raw:
        raw_adata = ad.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
        var_to_keep = raw_adata.var.index[~raw_adata.var.index.isin(deprecated)].tolist()
        raw_adata = raw_adata[:, var_to_keep]
        adata.raw = raw_adata
    return adata


def remap_deprecated_features(*, adata: ad.AnnData, remapped_features: Dict[str, str]) -> ad.AnnData:
    # Use remapped_terms to map to the new term ids
    adata.var.index = [remapped_features.get(val, val) for val in adata.var.index]

    # Repeat much of the same steps for the raw.var, if it exists
    if adata.raw:
        raw_adata = ad.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
        raw_adata.var.index = [remapped_features.get(val, val) for val in raw_adata.var.index]
        adata.raw = raw_adata
    return adata


def get_matrix_format(adata: ad.AnnData, matrix: Union[np.ndarray, sparse.spmatrix]) -> str:
    """
    Given a matrix, returns the format as one of: csc, csr, coo, dense
    or unknown.

    This mimics the scipy.sparse `format` property, but extends it to
    support ndarray and other classes AnnData may proxy the matrix with.
    """

    # Note: the AnnData proxy classes DO support the `format_str` property, but
    # doing a slice seemed safer, if less performant.  Using `format_str`, which
    # currently works, uses private API:
    #
    # >>> return getattr(matrix, "format_str", "dense)
    #
    matrix_format = "unknown"
    if adata.n_obs == 0 or adata.n_vars == 0:
        matrix_format = "dense"
    else:
        matrix_slice = matrix[0:1, 0:1]
        if isinstance(matrix_slice, sparse.spmatrix):
            matrix_format = matrix_slice.format
        elif isinstance(matrix_slice, np.ndarray):
            matrix_format = "dense"

    assert matrix_format in ["unknown", "csr", "csc", "coo", "dense"]
    return matrix_format


def getattr_anndata(adata: ad.AnnData, attr: str = None):
    """
    same as getattr but handles the special case of "raw.var" for an anndata.AndData object

    :param anndata.AnnData adata: the anndata.AnnData object from which to extract an attribute
    :param str attr: name of the attribute to extract

    :return the attribute or none if it does not exist
    """

    if attr == "raw.var":
        if adata.raw:
            return adata.raw.var
        else:
            return None
    else:
        return getattr(adata, attr)


def read_h5ad(h5ad_path: Union[str, bytes, os.PathLike]) -> ad.AnnData:
    """
    Reads h5ad into adata
    :params Union[str, bytes, os.PathLike] h5ad_path: path to h5ad to read

    :rtype None
    """
    try:
        adata = ad.read_h5ad(h5ad_path, backed="r")

        # This code, and AnnData in general, is optimized for row access.
        # Running backed, with CSC, is prohibitively slow. Read the entire
        # AnnData into memory if it is CSC.
        if (get_matrix_format(adata, adata.X) == "csc") or (
            (adata.raw is not None) and (get_matrix_format(adata, adata.raw.X) == "csc")
        ):
            logger.warning("Matrices are in CSC format; loading entire dataset into memory.")
            adata = adata.to_memory()

    except (OSError, TypeError):
        logger.info(f"Unable to open '{h5ad_path}' with AnnData")
        sys.exit(1)

    return adata


def get_hash_digest_column(dataframe):
    """
    Get column with hash digest for each row in dataframe.
    """
    df_index = dataframe.index.to_series()
    assert df_index.is_unique
    return (
        df_index.map(xxh3_64_intdigest)
        .astype(np.uint64)
        .apply(lambda v: b85encode(v.to_bytes(8, "big")).decode("ascii"))
    )



import json
import os

import anndata as ad

from . import utils

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# fmt: off
# ONTOLOGY TERMS TO UPDATE ACROSS ALL DATASETS IN CORPUS
# Initialization is AUTOMATED for newly deprecated terms that have 'Replaced By' terms in their ontology files

# Curators should review the monthly 'Curator Report' and add deprecated term replacements to corresponding map if
# 'Replaced By' is not available for a deprecated term.

# If Curators have non-deprecated term changes to apply to all datasets in the corpus where applicable,
# add them here.

# We have used this mapping for automigrating terms, not sure if this is accounted for in other migration logic
with open(os.path.join(BASE_DIR, "migrate_files/automigrate_terms.json"), "r") as file:
    DEV_STAGE_AUTO_MIGRATE_MAP = json.load(file)

ONTOLOGY_TERM_MAPS = {
    "assay": {
    },
    "cell_type": {
        "CL:4023070": "CL:4023064", # AUTOMATED
    },
    "development_stage": DEV_STAGE_AUTO_MIGRATE_MAP,
    "disease": {
    },
    "organism": {
    },
    "self_reported_ethnicity": {
    },
    "sex": {
    },
    "tissue": {
    },
}

DEPRECATED_FEATURE_IDS = [
]

# Dictionary for CURATOR-DEFINED remapping of deprecated feature IDs, if any, to new feature IDs.
GENCODE_MAPPER = {}

with open(os.path.join(BASE_DIR, "migrate_files/donor_updates.json"), "r") as file:
    DONOR_DEV_STAGE_MAP = json.load(file)

with open(os.path.join(BASE_DIR, "migrate_files/title_donor_updates.json"), "r") as file:
    TITLE_DONOR_DEV_STAGE_MAP = json.load(file)
# fmt: on


def migrate(input_file, output_file, collection_id, dataset_id):
    print(f"Converting {input_file} into {output_file}")

    dataset = ad.read_h5ad(input_file, backed="r")
    if dataset.raw is not None and DEPRECATED_FEATURE_IDS:
        dataset = dataset.to_memory()

    # AUTOMATED, DO NOT CHANGE
    for ontology_name, deprecated_term_map in ONTOLOGY_TERM_MAPS.items():
        utils.replace_ontology_term(dataset.obs, ontology_name, deprecated_term_map)

    # CURATOR-DEFINED, DATASET-SPECIFIC UPDATES
    # Use the template below to define dataset and collection specific ontology changes. Will only apply to dataset
    # if it matches a condition.
    # If no such changes are needed, leave blank
    # Examples:
    # if dataset_id == "<dataset_1_id>":
    #   <no further logic necessary>
    #   utils.replace_ontology_term(df, <ontology_name>, {"term_to_replace": "replacement_term", ...})
    # elif dataset_id == "<dataset_2_id>":
    #   <custom transformation logic beyond scope of util functions>
    # elif collection_id == "<collection_1_id>":
    #   <no further logic necessary>
    #   utils.replace_ontology_term(df, <ontology_name>, {"term_to_replace": "replacement_term", ...})
    # elif collection_id == "<collection_2_id>":
    #   <custom transformation logic beyond scope of replace_ontology_term>
    # ...

    # logic for mapping donor updates
    if collection_id in DONOR_DEV_STAGE_MAP:
        collection_donor_dev_map = DONOR_DEV_STAGE_MAP[collection_id]
        utils.map_ontology_term(dataset.obs, "development_stage", "donor_id", collection_donor_dev_map)

    # private dataset titles in title_donor_updates.json should be unique in corpus
    if dataset.uns["title"] in TITLE_DONOR_DEV_STAGE_MAP:
        dataset_donor_dev_map = TITLE_DONOR_DEV_STAGE_MAP[dataset.uns["title"]]
        utils.map_ontology_term(dataset.obs, "development_stage", "donor_id", dataset_donor_dev_map)

    # https://github.com/chanzuckerberg/single-cell-curation/issues/958

    if collection_id == "c114c20f-1ef4-49a5-9c2e-d965787fb90c":
        utils.replace_ontology_term(dataset.obs, "assay", {"EFO:0010550": "EFO:0030028"})

    if collection_id == "45d5d2c3-bc28-4814-aed6-0bb6f0e11c82":
        utils.replace_ontology_term(dataset.obs, "assay", {"EFO:0010550": "EFO:0030028"})

    if collection_id == "962df42d-9675-4d05-bc75-597ec7bf4afb":
        utils.replace_ontology_term(dataset.obs, "development_stage", {"unknown": "HsapDv:0000264"})

    if collection_id == "48d354f5-a5ca-4f35-a3bb-fa3687502252":
        utils.map_ontology_term(dataset.obs, "development_stage", "var_time", {"P7": "MmusDv:0000117"})

    if collection_id == "613f5480-4957-4f80-b804-0e2b85ac454c":
        utils.map_ontology_term(
            dataset.obs,
            "development_stage",
            "age",
            {
                "P4": "MmusDv:0000114",
                "P7": "MmusDv:0000117",
            },
        )

    if collection_id == "d86517f0-fa7e-4266-b82e-a521350d6d36":
        utils.map_ontology_term(
            dataset.obs,
            "development_stage",
            "Dataset",
            {
                "RomanovDev10x": "MmusDv:0000144",
                "Mickelsen10x": "MmusDv:0000149",
                "Flynn10x": "MmusDv:0000149",
            },
        )
        utils.map_ontology_term(
            dataset.obs,
            "development_stage",
            "SRA_ID",
            {
                "SRR5164436": "MmusDv:0000178",
                "SRR5164437": "MmusDv:0000178",
                "SRR5164438": "MmusDv:0000149",
            },
        )

    # dev migration revealed column 'feature_type' in var and raw.var, titles should be unique in corpus
    titles_with_feature_type = [
        "Healthy reference",
        "Extended - Neural cells",
        "Extended - Endothelial cells",
        "Extended+ - 18485 genes",
        "Healthy - Small intestine (adult/pediatric)",
        "Healthy - Mesenchymal (adult/pediatric)",
        "Healthy - Stomach",
        "Healthy - Large Intestine (adult/pediatric)",
        "Healthy - Mesenchymal (first trimester)",
        "Extended - All genes",
        "Healthy - Small Intestine (first trimester)",
        "Healthy - Neural cells",
        "Healthy - T and NK cells",
        "Healthy - Myeloid cells",
        "Extended - Small intestine (adult/pediatric)",
        "Extended - T and NK cells",
        "Healthy - Salivary gland",
        "Extended - Large Intestine (adult/pediatric)",
        "Healthy - Large Intestine (first trimester)",
        "Healthy - Oesophagus",
        "Extended - Stomach",
        "Healthy - Mesenchymal (second trimester)",
        "Extended - Small intestine (adult/pediatric)- 18485genes",
        "Healthy - Oral mucosa",
        "Extended - B and B plasma cells",
        "Extended - Myeloid_with_neutrophils",
        "Healthy - B and B plasma cells",
        "Healthy - Small Intestine (second trimester)",
        "Healthy - Endothelial cells",
        "Extended - Myeloid cells",
        "Extended - Mesenchymal (adult/pediatric)",
        "Healthy - Large Intestine (second trimester)",
    ]

    if dataset.uns["title"] in titles_with_feature_type or collection_id == "e5f58829-1a66-40b5-a624-9046778e74f5":
        if dataset.raw:
            dataset.raw.var.drop(columns="feature_type", inplace=True)

        dataset.var.drop(columns="feature_type", inplace=True)

    if GENCODE_MAPPER:
        dataset = utils.remap_deprecated_features(adata=dataset, remapped_features=GENCODE_MAPPER)

    # AUTOMATED, DO NOT CHANGE -- IF GENCODE UPDATED, DEPRECATED FEATURE FILTERING ALGORITHM WILL GO HERE.
    if DEPRECATED_FEATURE_IDS:
        dataset = utils.remove_deprecated_features(adata=dataset, deprecated=DEPRECATED_FEATURE_IDS)

    dataset.write(output_file, compression="gzip")



from setuptools import setup

with open("requirements.txt") as fh:
    requirements = fh.read().splitlines()

setup(
    name="cellxgene-schema",
    version="5.2.1",
    url="https://github.com/chanzuckerberg/single-cell-curation",
    license="MIT",
    author="Chan Zuckerberg Initiative",
    author_email="cellxgene@chanzuckerberg.com",
    description="Tool for applying and validating cellxgene integration schema to single cell datasets",
    long_description="Tool for applying and validating cellxgene integration schema to single cell datasets",
    install_requires=requirements,
    python_requires=">=3.10",
    packages=["cellxgene_schema"],
    package_dir={"cellxgene_schema": "cellxgene_schema"},
    package_data={"cellxgene_schema": ["gencode_files/*gz", "migrate_files/*json", "schema_definitions/*yaml"]},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={"console_scripts": ["cellxgene-schema = cellxgene_schema.cli:schema_cli"]},
)



import sys

import click


@click.group(
    name="schema",
    subcommand_metavar="COMMAND <args>",
    short_help="Apply and validate the cellxgene data integration schema to an h5ad file.",
    context_settings=dict(max_content_width=85, help_option_names=["-h", "--help"]),
)
def schema_cli():
    pass


@click.command(
    name="validate",
    short_help="Check that an h5ad follows the cellxgene data integration schema.",
    help="Check that an h5ad follows the cellxgene data integration schema. If validation fails this command will "
    "return an exit status of 1 otherwise 0. When the '--add-labels <FILE>' tag is present, the command will add "
    "ontology/gene labels based on IDs and write them to a new h5ad.",
)
@click.argument("h5ad_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.option(
    "-a",
    "--add-labels",
    "add_labels_file",
    help="When present it will add labels to genes and ontologies based on IDs",
    required=False,
    default=None,
    type=click.Path(exists=False, dir_okay=False, writable=True),
)
@click.option("-i", "--ignore-labels", help="Ignore ontology labels when validating", is_flag=True)
@click.option("-v", "--verbose", help="When present will set logging level to debug", is_flag=True)
def schema_validate(h5ad_file, add_labels_file, ignore_labels, verbose):
    # Imports are very slow so we defer loading until Click arg validation has passed

    print("Loading dependencies")
    try:
        import anndata  # noqa: F401
    except ImportError:
        raise click.ClickException("[cellxgene] cellxgene-schema requires anndata") from None

    print("Loading validator modules")
    from .validate import validate

    is_valid, _, _ = validate(h5ad_file, add_labels_file, ignore_labels=ignore_labels, verbose=verbose)
    if is_valid:
        sys.exit(0)
    else:
        sys.exit(1)


@click.command(
    name="remove-labels",
    short_help="Create a copy of an h5ad without portal-added labels",
    help="Create a copy of an h5ad without portal-added labels.",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
def remove_labels(input_file, output_file):
    from .remove_labels import AnnDataLabelRemover

    print("Loading dependencies")
    try:
        import anndata  # noqa: F401
    except ImportError:
        raise click.ClickException("[cellxgene] cellxgene-schema requires anndata") from None

    print(f"Loading h5ad from {input_file}")
    adata = anndata.read_h5ad(input_file)
    anndata_label_remover = AnnDataLabelRemover(adata)
    if not anndata_label_remover.schema_def:
        return
    print("Removing labels")
    anndata_label_remover.remove_labels()
    print(f"Labels have been removed. Writing to {output_file}")
    anndata_label_remover.adata.write(output_file, compression="gzip")


@click.command(
    name="migrate",
    short_help="Convert an h5ad to the latest schema version.",
    help="Convert an h5ad from the previous to latest minor schema version. No validation will be "
    "performed on either the input or the output file.",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
@click.option("--collection_id", default=None, type=str, help="Collection ID, if migrating already published dataset")
@click.option("--dataset_id", default=None, type=str, help="Dataset ID, if migrating already published dataset")
def migrate(input_file, output_file, collection_id, dataset_id):
    from .migrate import migrate

    migrate(input_file, output_file, collection_id, dataset_id)


schema_cli.add_command(schema_validate)
schema_cli.add_command(migrate)
schema_cli.add_command(remove_labels)

if __name__ == "__main__":
    schema_cli()
